#!/usr/bin/perl -w

######################################################
#
#	iso wrapper exploration
#
#	The iso wrapper has the ability to produce what 
#	should be the value for a given input parameter 'y'
#	while sweeping through a range of values for
#	another parameter 'x' to maintain constant a 
#	given output parameter.
#
#	The algorithm of this wrapper is based on
#	a binary search
#
#	The results are printed in a file <output_file_name>.txt
#	and a graph can be plot in a file <output_file_name>.eps
#
#	How to use it:
#	./iso.pl	<dc config file>								datacenter configuration file
#				<resource config file>							resource configuration file
#				<output_file_name>								filename of the output file without extension (the extensions are .txt and .eps)
#				<fix_output_name[:reference value]>				output parameter to maintain constant
#																optional: the reference value is optional if you want provide a value comming from another configuration
#																in that case you can put 1 to <y_parameter_value_ref> to obtain absolute values of the y_parameter_name in your iso experiment
#				<precision>										precision of the binary search algorithm
#				<x_parameter_name>								parameter x name 
#				<x_parameter_value_ref>							reference value of x 
#				<x_parameter_range_min>							minimale value of the x range
#				<x_parameter_range_max>							maximale value of the x range
#				<x_parameter_range_step>						number of steps in the x range
#				<x_parameter_range_type: absolute/relative>		the range can be absolute or relative
#				<y_parameter_name>								parameter y name
#				<y_parameter_value_ref>							reference value of y
#				[min:y_parameter_range_min]						optional: minimale value of y to guide the binary search algorithm if not specified, default values are assumed
#																usage: min:value
#				[max:y_parameter_range_max]						optional: maximale value of y to guide the binary search algorithm if not specified, default values are assumed
#																usage: max:value
#				[xaxis:name]									optional to put a name 'name' for the x-axis of the graph (by default <x_parameter_name> is used)
#				[yaxis:name]									optional to put a name 'name' for the y-axis of the graph (by default <y_parameter_name> is used)
#				[iso:name]										optional to put a name 'name' for the iso curve of the graph (by default iso <fix_output_name> is used)
#
#				[nognuplot]										optional: put "nognuplot" if you do not want to get the graph
#				[getresult]										optional: put "getresult" to print the last value found by the iso exploration
#																If no value was found, the value is set to -1
#																(to be used in other wrappers calling the iso wrapper)
#																
#
#	Remark:
#		In case the iso wrapper doesn't find a valid solution you get 'x' in the output file
#
#
#	Examples of command:
#
#	./iso.pl ../template/dc_template.params ../template/r_template.params iso_tco tco 0.0001 pvf 0 0 0.1 0.01 absolute c_chip 100 xaxis:"PVF" yaxis:"Normalized Processor Price" iso:"iso TCO"
#
#   ./iso.pl ../template/dc2_template.params ../template/r2_template.params iso_perftco "PerformanceOverTCO" 0.0001 pvf 0 0 0.1 0.01 absolute c_chip 100 xaxis:"PVF" yaxis:"Normalized Processor Price" iso:"iso Perf/TCO"
#
######################################################

use strict;

use File::Copy;

use lib "../lib/"; 
use helper;

###########################################################
#
#	check valid input
#
###########################################################
my $argv_size = scalar @ARGV;
if ($argv_size < 13)
{
	print "usage: ./iso.pl <dc config file> <resource config file> <output_file_name> <fix_output_name[:reference value]> <precision> <x_parameter_name> <x_parameter_value_ref> <x_parameter_range_min> <x_parameter_range_max> <x_parameter_range_step> <x_parameter_range_type: absolute/relative> <y_parameter_name> <y_parameter_value_ref> [min:y_parameter_range_min] [max:y_parameter_range_max] [nognuplot] [getresult]\n";
	exit(0);
}


my $dc_filename=$ARGV[0];
my $resource_filename=$ARGV[1];

my $filename=$ARGV[2];

my $constant_output_value_name=$ARGV[3];

my $cst_defined=0;
my $value_ref=0;
my @tmp=split(':' ,$constant_output_value_name);
my $tmp_size=scalar @tmp;
if($tmp_size==2)
{
	$constant_output_value_name=$tmp[0];
	$value_ref=$tmp[1];
	$cst_defined=1;
}
elsif($tmp_size==1){;}
else{print("error: $constant_output_value_name not correctly defined\n");exit(0);}

my $precision=$ARGV[4];

my $x_parameter_name=$ARGV[5];
my $x_parameter_value_ref=$ARGV[6];
my $x_parameter_range_value_min=$ARGV[7];
my $x_parameter_range_value_max=$ARGV[8];
my $x_parameter_range_value_step=$ARGV[9];

my $x_range_relative;
if($ARGV[10] eq "relative")
{
	$x_range_relative=1;
}
elsif($ARGV[10] eq "absolute")
{
	$x_range_relative=0;
}
else{print("error: parameter has to be relative or absolute\n");exit(0);}

my $y_parameter_name=$ARGV[11];
my $y_parameter_value_ref=$ARGV[12];


my @range = determinedRange($y_parameter_name);
my $y_parameter_range_value_min= $range[0];
my $y_parameter_range_value_max= $range[1];
my $gnuplot=1;
my $printres=0;
my $x_axis_name=$x_parameter_name;
my $y_axis_name=$y_parameter_name;
my $iso_name="iso $constant_output_value_name";
for(my $i=13;$i<19;$i++)
{
	if(defined($ARGV[$i]))
	{
		my $arg_tmp=$ARGV[$i];
		
		my @list=split(':',$arg_tmp);
		
		if( $list[0] eq "nognuplot")
		{
			$gnuplot=0;
		}
		elsif($list[0] eq "getresult")
		{
			$printres=1;
		}
		elsif($list[0] eq "min")
		{
			if(scalar(@list)!=2){print ("Error: parameter $arg_tmp not correctly defined\n");exit(0);}
			$y_parameter_range_value_min=$list[1];
		}
		elsif($list[0] eq "max")
		{
			if(scalar(@list)!=2){print ("Error: parameter $arg_tmp not correctly defined\n");exit(0);}
			$y_parameter_range_value_max=$list[1];
		}
		elsif($list[0] eq "xaxis")
		{
			$x_axis_name=$list[1];
		}
		elsif($list[0] eq "yaxis")
		{
			if(scalar(@list)!=2){print ("Error: parameter $arg_tmp not correctly defined\n");exit(0);}
			$y_axis_name=$list[1];
		}elsif($list[0] eq "iso")
		{
			if(scalar(@list)!=2){print ("Error: parameter $arg_tmp not correctly defined\n");exit(0);}
			$iso_name=$list[1];
		}
		else{print ("Error: parameter $arg_tmp not correctly defined\n");exit(0);}
	}
	else{last;}
}



######################################################
#
# General Initialization
#
######################################################

#create the helper
my $helper = helper->new("");

#check that the inputs are correct
if($helper->checkInputParameterName($x_parameter_name)==-1){print("Execution stopped\n");exit(0); }
if($helper->checkInputParameterName($y_parameter_name)==-1){print("Execution stopped\n");exit(0); }
#check that the output name is correct
if($helper->checkOutputValueName($constant_output_value_name)==-1){print("Execution stopped\n");exit(0); }
if($helper->checkOutputValueName($constant_output_value_name)==1){print("Warning: The parameter 'report_resources' in the DC file must be set to 1 to get correct results\n"); }

#to get the current dir
my $currentdir=`pwd`;
chomp $currentdir;

#open outputfile
open(OUTPUTFILE,">$currentdir/output/$filename.txt") || die ("Error when opening $currentdir/output/$filename.txt") ;

######################################################
#
# Specific Initialization
#
######################################################


#return value of the iso exploration (to be used in other wrappers calling the iso wrapper)
my $return_value=-1;

#get the parameters from files
my $ptr_dc=$helper->parserDCFile("$dc_filename");
my %dc=%{$ptr_dc};

my $ptr_resource=$helper->parserResourceFile("$resource_filename");
my %resource=%{$ptr_resource};


	####################################################
	#get the reference value 
	#(if needed i.e. $cst_defined==0)
	####################################################

	if($cst_defined==0)
	{
		#generate the current configuration file
		$dc{"$x_parameter_name"}=$x_parameter_value_ref;
		$dc{"$y_parameter_name"}=$y_parameter_value_ref;
		$helper->createParameterFile("$currentdir/dc.params",\%dc);
	
		$resource{"$x_parameter_name"}=$x_parameter_value_ref;
		$resource{"$y_parameter_name"}=$y_parameter_value_ref;
		$helper->createParameterFile("$currentdir/s.params",\%resource);
	
		#call the eetco kernel and keep the output
		$helper->callKernel("$currentdir/dc.params","$currentdir/s.params","$currentdir/output/ref");
	
		#get the result value
		my $ptr_res=$helper->parserResultKernel("$currentdir/output/ref");
		my %res=%{$ptr_res};
		$value_ref=$res{"$constant_output_value_name"};
	}


####################################################
# main exploration loop
####################################################

#exploration loop of x
for(my $x=$x_parameter_range_value_min;$x<=$x_parameter_range_value_max;$x+=$x_parameter_range_value_step)
{	
	$return_value=-1;
	
	my $x_value=$x;

	if($x_range_relative==1)
	{
		$x_value=$x * $x_parameter_value_ref;
	}
		
	#######################################################
	# check if there is a solution in the ymin ; ymax range
	#######################################################
	
		#######################################################
		#get the value with y min
		#######################################################
	
	#create the DC configuration file
	$dc{$x_parameter_name}=$x_value;
	$dc{$y_parameter_name}=$y_parameter_range_value_min;
	$helper->createParameterFile("$currentdir/dc.params",\%dc);
	
	#create the resource configuration file
	$resource{$x_parameter_name}=$x_value;
	$resource{$y_parameter_name}=$y_parameter_range_value_min;
	$helper->createParameterFile("$currentdir/s.params",\%resource);
	
	#call the eetco kernel and keep the output
	$helper->callKernel("$currentdir/dc.params","$currentdir/s.params","$currentdir/output/min");
		
	#get the value
	my $ptr_res=$helper->parserResultKernel("$currentdir/output/min");
	my %res=%{$ptr_res};
	my $value_with_ymin=$res{$constant_output_value_name};
	#print "$constant_output_value_name";
	my $normalized_with_ymin=$value_with_ymin/$value_ref;
	
		#######################################################
		#get the value with y max
		#######################################################
	
	#create the DC configuration file
	$dc{$x_parameter_name}=$x_value;
	$dc{$y_parameter_name}=$y_parameter_range_value_max;
	$helper->createParameterFile("$currentdir/dc.params",\%dc);
	
	#create the resource configuration file
	$resource{$x_parameter_name}=$x_value;
	$resource{$y_parameter_name}=$y_parameter_range_value_max;
	$helper->createParameterFile("$currentdir/s.params",\%resource);
	
	#call the eetco kernel and keep the output
	$helper->callKernel("$currentdir/dc.params","$currentdir/s.params","$currentdir/output/max");
	
	#get the value
	$ptr_res=$helper->parserResultKernel("$currentdir/output/max");
	%res=%{$ptr_res};
	my $value_with_ymax=$res{$constant_output_value_name};
	
	my $normalized_with_ymax=$value_with_ymax/$value_ref;
	
	#get the caracteristics of the function
	my $positive_y_function;
	if($value_with_ymin<$value_with_ymax)
	{
		$positive_y_function=1;
	}
	elsif($value_with_ymin>$value_with_ymax)
	{
		$positive_y_function=0;
	}
	
	#######################################################
	# no solution represented by x in the output file
	#######################################################
	if( ($normalized_with_ymax  > 1+$precision && $normalized_with_ymin  > 1+$precision) || ($normalized_with_ymax  < 1-$precision && $normalized_with_ymin  < 1-$precision) )
	{
		print(OUTPUTFILE "$x\tx\n");
	}

	#######################################################
	#solution exist: perform a binary search
	#######################################################
	else
	{
		my $normalized_value=0;
		
		my $a=$y_parameter_range_value_min;
		my $b=$y_parameter_range_value_max;
		
		my $y_current=0;
		
		while(  !( (1 <= $normalized_value + $precision) && (1>=$normalized_value - $precision) ) )
		{
			$y_current= ($a+$b)/2;
			
			#create the DC configuration file
			$dc{$x_parameter_name}=$x_value;
			$dc{$y_parameter_name}=$y_current;
			$helper->createParameterFile("$currentdir/dc.params",\%dc);
			
			#create the resource configuration file
			$resource{$x_parameter_name}=$x_value;
			$resource{$y_parameter_name}=$y_current;
			$helper->createParameterFile("$currentdir/s.params",\%resource);
			
			#call the eetco kernel and keep the output
			$helper->callKernel("$currentdir/dc.params","$currentdir/s.params","$currentdir/output/x_$x_value._y_$y_current");
			
			#get the value
			$ptr_res=$helper->parserResultKernel("$currentdir/output/x_$x_value._y_$y_current");
			%res=%{$ptr_res};
			my $current_value=$res{$constant_output_value_name};
			
			$normalized_value=$current_value/$value_ref;
			
			if($positive_y_function==1)
			{
				if($normalized_value -1 >0)
				{
					$b=$y_current;
				}
				if($normalized_value -1 <0)
				{
					$a=$y_current;
				}
			}
			else
			{				
				if($normalized_value -1 >0)
				{
					$a=$y_current;
				}
				if($normalized_value -1 <0)
				{
					$b=$y_current;
				}
			}
		}
		
		my $y_value=$y_current/$y_parameter_value_ref;

		my $MTTFall = 0;
		
		if ($resource{"chip_mttf"} && $resource{"mttf_switchport"} && $resource{"mttf_disk"} && $resource{"mttf_dram"})
		{
			$MTTFall = 1/ ( ($resource{k_chipspersrvmodule}/$resource{chip_mttf})+($resource{n_dram}/$resource{mttf_dram})+($resource{n_disk}/$resource{mttf_disk})+($resource{n_switchport}/$resource{mttf_switchport}) );
		}
		
		print(OUTPUTFILE "$x\t$y_value\t$y_current\t" . $res{"co2peryr"} . "\t" . $res{"n_coldsp"} . "\t" . $res{"n_srvmodules"} . "\t" . $MTTFall . "\n");
		$return_value=$y_value;
	}
}

close(OUTPUTFILE);


#keep the output file in the current directory
copy("$currentdir/output/$filename.txt","$currentdir/$filename.txt") or die "Copy failed: $!";

####################################################
# gnulpot call
####################################################
if($gnuplot==1)
{
	my @curves_names=("$iso_name");
	$helper->makeGnuplotCurves2D("$currentdir/$filename.txt","$currentdir","$filename","$x_axis_name","$y_axis_name",\@curves_names);
}

####################################################
# clean
####################################################
$helper->deleteFile("$currentdir/dc.params");
$helper->deleteFile("$currentdir/s.params");
$helper->deleteFileInDir("$currentdir/output/");

####################################################
# return the last found value of the iso exploration
# (-1 if no solution was found)
####################################################
if($printres==1)
{
	print("$return_value\n");
}

###################################################
###################################################
###################################################
###################################################
#Function

sub determinedRange#($)
{
	my $parameter_name=$_[0];

		my %parameters_range=
	(
	
	##DC parameters
	"c_buildingpersqm" , [1 , 50000],
	"a_coolingarea" , [1 , 10],
	"dcdepr" , [1 , 50],
	"c_coolpowerdeplperW" , [1 , 1000],
	"c_electperKWH" , [0 , 10],
	"PUE" , [1 , 10],
	"salary_perrackpermonth" , [0 , 1000],
	#"report_resources" , [min , max],
	"n_othermodules" , [0 , 500000],
	"p_othermodules" , [0 , 2000],
	"p_othermodulesidle" , [0 , 2000],
	"n_othermodulesrepl" , [0 , 500000],
	"c_nonsrvmodule" , [0 , 10000],
	"depr_othermodules" , [1 , 50],
	"u_othermodules" , [0 , 1],
	"a_othermodules" , [0 , 10000],
	"n_racks_othermodules" , [0 , 10000],
	"cost_of_money", [0,1],
	
	##Resource parameters
	"n_srvmodulesreq" , [1 , 500000],
	"u" , [0 , 1],
	"coef_maintenance" , [0 , 1],
	"k_bladesperrack" , [1 , 50],
	"k_modulesperblade" , [1 , 500],
	#"height" , [min , max],
	#"width" , [min , max],
	#"depth" , [min , max],
	#"minimumdistance" , [min , max],
	#"useddistance" , [min , max],
	"c_networkperrack" , [0 , 100000],
	"networkdepr" , [1 , 50],
	"network_peak_powerperrack" , [0 , 10000],
	"srvdepr" , [1 , 50],
	"mttr" , [1 , 1000],
	"mttf_coef" , [0.1 , 1],
	"nb_time_step" , [1 , 1095],
	"SPUE" , [1 , 10],
	"k_chipspersrvmodule" , [1 , 200],
	"c_system" , [0 , 50000],
	"p_system" , [0 , 1000],
	"p_system_idle" , [0 , 1000],
	"n_dram" , [0 , 50],
	"c_dram" , [0 , 1000],
	"p_dram" , [0 , 1000],
	"p_dram_idle" , [0 , 1000],
	"mttf_dram" , [50 , 1000],
	"n_disk" , [0 , 50],
	"c_disk" , [0 , 1000],
	"p_disk" , [0 , 1000],
	"p_disk_idle" , [0 , 1000],
	"mttf_disk" , [50 , 1000],
	"n_switchport" , [0 , 50],
	"c_switchport" , [0 , 1000],
	"p_switchport" , [0 , 1000],
	"p_switchport_idle" , [0 , 1000],
	"mttf_switchport" , [50 , 1000],
	"c_other" , [0 , 50000],
	"p_other" , [0 , 1000],
	"p_other_idle" , [0 , 1000],
	"c_chip" , [0 , 10000],
	"pvf" , [0 , 0.9],
	"p_chip" , [0 , 1000],
	"p_chip_idle" , [0 , 1000],
	"chip_mttf" , [15 , 1000],
	"chip_perf" , [0 , 1000],
	"co2peryr" , [15000000 , 20000000],	
	);
	
	
	#check if the parameter is present in %parameters_range
	if(!defined($parameters_range{$parameter_name}))
	{
		print("No range defined for '$parameter_name', sorry\n");exit(0);
	}
	
	my @res=@{$parameters_range{$parameter_name}};
	
	print("Assumed range for '$parameter_name' [$res[0];$res[1]]\n");
	
	return @res;

}
